@@ -4,6 +4,7 @@
  *
  * Copyright (C) 2002-2006 Jody Goldberg (jody@gnome.org)
  *                    	   Tambet Ingo   (tambet@ximian.com)
+ * Copyright (C) 2014 Morten Welinder (terra@gnome.org)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of version 2.1 of the GNU Lesser General Public
@@ -42,25 +43,26 @@ enum {
 static GObjectClass *parent_class;
 
 typedef struct {
-	guint16     entries;
-	guint32     dir_pos;
-	GList	   *dirent_list;
+	guint32     entries;
+	gsf_off_t   dir_pos;
+	GPtrArray  *dirents;
 	GsfZipVDir *vdir;
 
 	int ref_count;
 } ZipInfo;
 
 struct _GsfInfileZip {
-	GsfInfile parent;
+	GsfInfile base;
 
 	GsfInput  *source;
 	ZipInfo	  *info;
+	gboolean  zip64;
 
 	GsfZipVDir   *vdir;
 
 	z_stream  *stream;
-	guint32   restlen;
-	guint32   crestlen;
+	gsf_off_t restlen;
+	gsf_off_t crestlen;
 
 	guint8   *buf;
 	size_t    buf_size;
@@ -100,10 +102,10 @@ zip_make_modtime (guint32 dostime)
 static GsfZipVDir *
 vdir_child_by_name (GsfZipVDir *vdir, char const *name)
 {
-	GSList *l;
+	unsigned ui;
 
-	for (l = vdir->children; l; l = l->next) {
-		GsfZipVDir *child = (GsfZipVDir *) l->data;
+	for (ui = 0; ui < vdir->children->len; ui++) {
+		GsfZipVDir *child = g_ptr_array_index (vdir->children, ui);
 		if (strcmp (child->name, name) == 0)
 			return child;
 	}
@@ -113,7 +115,9 @@ vdir_child_by_name (GsfZipVDir *vdir, char const *name)
 static GsfZipVDir *
 vdir_child_by_index (GsfZipVDir *vdir, int target)
 {
-	return g_slist_nth_data (vdir->children, target);
+	return (unsigned)target < vdir->children->len
+		? g_ptr_array_index (vdir->children, target)
+		: NULL;
 }
 
 static void
@@ -143,16 +147,15 @@ vdir_insert (GsfZipVDir *vdir, char const * name, GsfZipDirent *dirent)
 }
 
 static gsf_off_t
-zip_find_trailer (GsfInfileZip *zip)
+zip_find_trailer (GsfInfileZip *zip, guint32 sig, gssize size)
 {
-	static guint8 const trailer_signature[] =
-		{ 'P', 'K', 0x05, 0x06 };
 	gsf_off_t offset, trailer_offset, filesize;
 	gsf_off_t maplen;
 	guint8 const *data;
+	guchar sig1 = sig & 0xff;
 
 	filesize = gsf_input_size (zip->source);
-	if (filesize < ZIP_TRAILER_SIZE)
+	if (filesize < size)
 		return -1;
 
 	trailer_offset = filesize;
@@ -173,9 +176,9 @@ zip_find_trailer (GsfInfileZip *zip)
 		p = (guchar *) data;
 
 		for (s = p + maplen - 1; (s >= p); s--, trailer_offset--) {
-			if ((*s == 'P') &&
-			    (p + maplen - 1 - s > ZIP_TRAILER_SIZE - 2) &&
-			    !memcmp (s, trailer_signature, sizeof (trailer_signature))) {
+			if (*s == sig1 &&
+			    p + maplen - 1 - s > size - 2 &&
+			    GSF_LE_GET_GUINT32 (s) == sig) {
 				return --trailer_offset;
 			}
 		}
@@ -202,27 +205,77 @@ zip_find_trailer (GsfInfileZip *zip)
 	return -1;
 }
 
+static guint8 const *
+zip_dirent_extra_field (guint8 const *extra, size_t elen,
+			guint16 typ, guint32 *pflen)
+{
+	while (TRUE) {
+		guint16 ftyp, flen;
+
+		if (elen == 0) {
+			*pflen = 0;
+			return NULL;
+		}
+
+		if (elen < 4)
+			goto bad;
+
+		ftyp = GSF_LE_GET_GUINT16 (extra);
+		flen = GSF_LE_GET_GUINT16 (extra + 2);
+		if (flen > elen - 4)
+			goto bad;
+
+		extra += 4;
+		elen -= 4;
+		if (ftyp == typ) {
+			/* Found the extended data.  */
+			*pflen = flen;
+			return extra;
+		}
+		extra += flen;
+		elen -= flen;
+	}
+
+bad:
+	*pflen = 0;
+	return NULL;
+}
+
+
 static GsfZipDirent *
 zip_dirent_new_in (GsfInfileZip *zip, gsf_off_t *offset)
 {
-	static guint8 const dirent_signature[] =
-		{ 'P', 'K', 0x01, 0x02 };
 	GsfZipDirent *dirent;
-	guint8 const *data;
-	guint16 name_len, extras_len, comment_len, compr_method, flags;
-	guint32 dostime, crc32, csize, usize, off;
+	guint8 const *data, *variable, *extra;
+	guint16 compr_method, flags;
+	guint32 dostime, crc32, disk_start, name_len, extras_len, comment_len, vlen, elen;
+	gsf_off_t off, csize, usize;
 	gchar *name;
+	guint8 header[ZIP_DIRENT_SIZE];
+	gboolean zip64;
 
-	/* Read data and check the header */
+	/* Read fixed-length part of data and check the header */
+	data = header;
 	if (gsf_input_seek (zip->source, *offset, G_SEEK_SET) ||
-	    NULL == (data = gsf_input_read (zip->source, ZIP_DIRENT_SIZE, NULL)) ||
-	    0 != memcmp (data, dirent_signature, sizeof (dirent_signature))) {
+	    !gsf_input_read (zip->source, ZIP_DIRENT_SIZE, header) ||
+	    GSF_LE_GET_GUINT32 (data) != ZIP_DIRENT_SIGNATURE) {
 		return NULL;
 	}
 
 	name_len =      GSF_LE_GET_GUINT16 (data + ZIP_DIRENT_NAME_SIZE);
 	extras_len =    GSF_LE_GET_GUINT16 (data + ZIP_DIRENT_EXTRAS_SIZE);
 	comment_len =   GSF_LE_GET_GUINT16 (data + ZIP_DIRENT_COMMENT_SIZE);
+	vlen = name_len + extras_len + comment_len;
+
+	/* Read variable part */
+	variable = gsf_input_read (zip->source, vlen, NULL);
+	if (!variable && vlen > 0)
+		return NULL;
+	if (FALSE && variable) gsf_mem_dump (variable, vlen);
+
+	extra = zip_dirent_extra_field (variable + name_len, extras_len,
+					ZIP_DIRENT_EXTRA_FIELD_ZIP64, &elen);
+	zip64 = (extra != NULL);
 
 	flags =         GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_FLAGS);
 	compr_method =  GSF_LE_GET_GUINT16 (data + ZIP_DIRENT_COMPR_METHOD);
@@ -231,12 +284,31 @@ zip_dirent_new_in (GsfInfileZip *zip, gsf_off_t *offset)
 	csize =         GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_CSIZE);
 	usize =         GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_USIZE);
 	off =           GSF_LE_GET_GUINT32 (data + ZIP_DIRENT_OFFSET);
+	disk_start =    GSF_LE_GET_GUINT16 (data + ZIP_DIRENT_DISKSTART);
 
-	if ((data = gsf_input_read (zip->source, name_len, NULL)) == NULL)
-		return NULL;
+	if (usize == 0xffffffffu && elen >= 8) {
+		usize = GSF_LE_GET_GUINT64 (extra);
+		extra += 8;
+		elen -= 8;
+	}
+	if (csize == 0xffffffffu && elen >= 8) {
+		csize = GSF_LE_GET_GUINT64 (extra);
+		extra += 8;
+		elen -= 8;
+	}
+	if (off == 0xffffffffu && elen >= 8) {
+		off = GSF_LE_GET_GUINT64 (extra);
+		extra += 8;
+		elen -= 8;
+	}
+	if (disk_start == 0xffffu && elen >= 4) {
+		disk_start = GSF_LE_GET_GUINT32 (extra);
+		extra += 4;
+		elen -= 4;
+	}
 
-	name = g_new (gchar, (gulong) (name_len + 1));
-	memcpy (name, data, name_len);
+	name = g_new (gchar, name_len + 1);
+	memcpy (name, variable, name_len);
 	name[name_len] = '\0';
 
 	dirent = gsf_zip_dirent_new ();
@@ -249,11 +321,12 @@ zip_dirent_new_in (GsfInfileZip *zip, gsf_off_t *offset)
 	dirent->usize =         usize;
 	dirent->offset =        off;
 	dirent->dostime =       dostime;
+	dirent->zip64 =         zip64;
 #if 0
 	g_print ("%s = 0x%x @ %" GSF_OFF_T_FORMAT "\n", name, off, *offset);
 #endif
 
-	*offset += ZIP_DIRENT_SIZE + name_len + extras_len + comment_len;
+	*offset += ZIP_DIRENT_SIZE + vlen;
 
 	return dirent;
 }
@@ -269,16 +342,17 @@ zip_info_ref (ZipInfo *info)
 static void
 zip_info_unref (ZipInfo *info)
 {
-	GList *p;
+	unsigned ui;
 
 	if (info->ref_count-- != 1)
 		return;
 
 	gsf_zip_vdir_free (info->vdir, FALSE);
-	for (p = info->dirent_list; p != NULL; p = p->next)
-		gsf_zip_dirent_free ((GsfZipDirent *) p->data);
-
-	g_list_free (info->dirent_list);
+	for (ui = 0; ui < info->dirents->len; ui++) {
+		GsfZipDirent *e = g_ptr_array_index (info->dirents, ui);
+		gsf_zip_dirent_free (e);
+	}
+	g_ptr_array_free (info->dirents, TRUE);
 
 	g_free (info);
 }
@@ -317,33 +391,54 @@ zip_dup (GsfInfileZip const *src, GError **err)
 static gboolean
 zip_read_dirents (GsfInfileZip *zip)
 {
-	guint8 const *trailer;
-	guint16 entries, i;
-	guint32 dir_pos;
+	guint8 const *data, *locator;
+	guint32 entries, i;
 	ZipInfo *info;
-	gsf_off_t offset;
+	gsf_off_t dir_pos, offset;
 
 	/* Find and check the trailing header */
-	offset = zip_find_trailer (zip);
-	if (offset < 0) {
-		zip->err = g_error_new (gsf_input_error_id (), 0,
-					_("No Zip trailer"));
-		return TRUE;
-	}
+	offset = zip_find_trailer (zip, ZIP_TRAILER_SIGNATURE, ZIP_TRAILER_SIZE);
+	if (offset < ZIP_ZIP64_LOCATOR_SIZE ||
+	    gsf_input_seek (zip->source, offset - ZIP_ZIP64_LOCATOR_SIZE, G_SEEK_SET))
+		goto bad;
 
-	if (gsf_input_seek (zip->source, offset, G_SEEK_SET) ||
-	    NULL == (trailer = gsf_input_read (zip->source, ZIP_TRAILER_SIZE, NULL))) {
-		zip->err = g_error_new (gsf_input_error_id (), 0,
-					_("Error reading Zip signature"));
-		return TRUE;
-	}
+	locator = gsf_input_read (zip->source, ZIP_TRAILER_SIZE + ZIP_ZIP64_LOCATOR_SIZE, NULL);
+	if (!locator)
+		goto bad;
+	data = locator + ZIP_ZIP64_LOCATOR_SIZE;
 
-	entries      = GSF_LE_GET_GUINT32 (trailer + ZIP_TRAILER_ENTRIES);
-	dir_pos      = GSF_LE_GET_GUINT32 (trailer + ZIP_TRAILER_DIR_POS);
+	entries      = GSF_LE_GET_GUINT16 (data + ZIP_TRAILER_ENTRIES);
+	dir_pos      = GSF_LE_GET_GUINT32 (data + ZIP_TRAILER_DIR_POS);
+
+	if (GSF_LE_GET_GUINT32 (locator) == ZIP_ZIP64_LOCATOR_SIGNATURE) {
+		guint32 disk, disks;
+		gsf_off_t zip64_eod_offset;
+
+		zip->zip64 = TRUE;
+
+		data = locator;
+		disk = GSF_LE_GET_GUINT32 (data + ZIP_ZIP64_LOCATOR_DISK);
+		zip64_eod_offset = GSF_LE_GET_GUINT64 (data + ZIP_ZIP64_LOCATOR_OFFSET);
+		disks = GSF_LE_GET_GUINT32 (data + ZIP_ZIP64_LOCATOR_DISKS);
+
+		if (disk != 0 || disks != 1)
+			goto bad;
+
+		if (gsf_input_seek (zip->source, zip64_eod_offset, G_SEEK_SET))
+			goto bad;
+		data = gsf_input_read (zip->source, ZIP_TRAILER64_SIZE, NULL);
+		if (!data ||
+		    GSF_LE_GET_GUINT32 (data) != ZIP_TRAILER64_SIGNATURE)
+			goto bad;
+
+		entries = GSF_LE_GET_GUINT64 (data + ZIP_TRAILER64_ENTRIES);
+		dir_pos = GSF_LE_GET_GUINT64 (data + ZIP_TRAILER64_DIR_POS);
+	}
 
 	info = g_new0 (ZipInfo, 1);
 	zip->info = info;
 
+	info->dirents = g_ptr_array_new ();
 	info->ref_count    = 1;
 	info->entries      = entries;
 	info->dir_pos      = dir_pos;
@@ -359,22 +454,26 @@ zip_read_dirents (GsfInfileZip *zip)
 			return TRUE;
 		}
 
-		info->dirent_list = g_list_append (info->dirent_list, d);
+		g_ptr_array_add (info->dirents, d);
 	}
 
 	return FALSE;
+
+bad:
+	zip->err = g_error_new (gsf_input_error_id (), 0,
+				_("Broken zip file structure"));
+	return TRUE;
 }
 
 static void
 zip_build_vdirs (GsfInfileZip *zip)
 {
-	GList *l;
-	GsfZipDirent *dirent;
+	unsigned ui;
 	ZipInfo *info = zip->info;
 
 	info->vdir = gsf_zip_vdir_new ("", TRUE, NULL);
-	for (l = info->dirent_list; l; l = l->next) {
-		dirent = (GsfZipDirent *) l->data;
+	for (ui = 0; ui < info->dirents->len; ui++) {
+		GsfZipDirent *dirent = g_ptr_array_index (info->dirents, ui);
 		vdir_insert (info->vdir, dirent->name, dirent);
 	}
 }
@@ -405,8 +504,6 @@ zip_init_info (GsfInfileZip *zip)
 static gboolean
 zip_child_init (GsfInfileZip *child, GError **errmsg)
 {
-	static guint8 const header_signature[] =
-		{ 'P', 'K', 0x03, 0x04 };
 	guint8 const *data = NULL;
 	guint16 name_len, extras_len;
 	const char *err = NULL;
@@ -415,18 +512,16 @@ zip_child_init (GsfInfileZip *child, GError **errmsg)
 
 	/* skip local header
 	 * should test tons of other info, but trust that those are correct
-	 **/
+	 */
 
-	if (gsf_input_seek (child->source, (gsf_off_t) dirent->offset, G_SEEK_SET))
+	if (gsf_input_seek (child->source, dirent->offset, G_SEEK_SET)) {
 		err = _("Error seeking to zip header");
-	else if (NULL == (data = gsf_input_read (child->source, ZIP_FILE_HEADER_SIZE, NULL)))
+	} else if (NULL == (data = gsf_input_read (child->source, ZIP_HEADER_SIZE, NULL)))
 		err = _("Error reading zip header");
-	else if (0 != memcmp (data, header_signature, sizeof (header_signature))) {
+	else if (GSF_LE_GET_GUINT32 (data) != ZIP_HEADER_SIGNATURE) {
 		err = _("Error incorrect zip header");
-		g_printerr ("Header is :\n");
-		gsf_mem_dump (data, sizeof (header_signature));
-		g_printerr ("Header should be :\n");
-		gsf_mem_dump (header_signature, sizeof (header_signature));
+		g_printerr ("Header is 0x%x\n", GSF_LE_GET_GUINT32 (data));
+		g_printerr ("Expected 0x%x\n", ZIP_HEADER_SIGNATURE);
 	}
 
 	if (NULL != err) {
@@ -438,10 +533,10 @@ zip_child_init (GsfInfileZip *child, GError **errmsg)
 	/* Throw clang a bone.  */
 	g_assert (data != NULL);
 
-	name_len =   GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_NAME_SIZE);
-	extras_len = GSF_LE_GET_GUINT16 (data + ZIP_FILE_HEADER_EXTRAS_SIZE);
+	name_len =   GSF_LE_GET_GUINT16 (data + ZIP_HEADER_NAME_SIZE);
+	extras_len = GSF_LE_GET_GUINT16 (data + ZIP_HEADER_EXTRAS_SIZE);
 
-	dirent->data_offset = dirent->offset + ZIP_FILE_HEADER_SIZE + name_len + extras_len;
+	dirent->data_offset = dirent->offset + ZIP_HEADER_SIZE + name_len + extras_len;
 	child->restlen  = dirent->usize;
 	child->crestlen = dirent->csize;
 
@@ -513,10 +608,10 @@ static guint8 const *
 gsf_infile_zip_read (GsfInput *input, size_t num_bytes, guint8 *buffer)
 {
 	GsfInfileZip *zip = GSF_INFILE_ZIP (input);
-	GsfZipVDir      *vdir = zip->vdir;
+	GsfZipVDir *vdir = zip->vdir;
 	gsf_off_t pos;
 
-	if (zip->restlen < num_bytes)
+	if (zip->restlen < (gsf_off_t)num_bytes)
 		return NULL;
 
 	switch (vdir->dirent->compr_method) {
@@ -542,7 +637,7 @@ gsf_infile_zip_read (GsfInput *input, size_t num_bytes, guint8 *buffer)
 
 		do {
 			int err;
-			int startlen;
+			gsf_off_t startlen;
 
 			if (zip->crestlen > 0 && zip->stream->avail_in == 0)
 				if (!zip_update_stream_in (zip))
@@ -614,7 +709,6 @@ gsf_infile_zip_seek (GsfInput *input, gsf_off_t offset, GSeekType whence)
 
 /*****************************************************************************/
 
-
 static GsfInput *
 gsf_infile_zip_new_child (GsfInfileZip *parent, GsfZipVDir *vdir, GError **err)
 {
@@ -631,8 +725,7 @@ gsf_infile_zip_new_child (GsfInfileZip *parent, GsfZipVDir *vdir, GError **err)
 	child->vdir = vdir;
 
 	if (dirent) {
-		gsf_input_set_size (GSF_INPUT (child),
-				    (gsf_off_t) dirent->usize);
+		gsf_input_set_size (GSF_INPUT (child), dirent->usize);
 
 		if (dirent->dostime) {
 			GDateTime *modtime = zip_make_modtime (dirent->dostime);
@@ -695,7 +788,7 @@ gsf_infile_zip_num_children (GsfInfile *infile)
 
 	if (!zip->vdir->is_directory)
 		return -1;
-	return g_slist_length (zip->vdir->children);
+	return zip->vdir->children->len;
 }
 
 static void
@@ -733,6 +826,7 @@ gsf_infile_zip_finalize (GObject *obj)
 	parent_class->finalize (obj);
 }
 
+
 static GObject*
 gsf_infile_zip_constructor (GType                  type,
 			    guint                  n_construct_properties,
@@ -747,6 +841,7 @@ gsf_infile_zip_constructor (GType                  type,
 		/* Special call from zip_dup.  */
 		zip->source = gsf_input_dup (zip->dup_parent->source, &zip->err);
 		zip->info = zip_info_ref (zip->dup_parent->info);
+		zip->zip64 = zip->dup_parent->zip64;
 		zip->dup_parent = NULL;
 	} else {
 		if (!zip_init_info (zip))
@@ -756,13 +851,13 @@ gsf_infile_zip_constructor (GType                  type,
 	return (GObject *)zip;
 }
 
-
 static void
 gsf_infile_zip_init (GObject *obj)
 {
 	GsfInfileZip *zip = (GsfInfileZip *)obj;
 	zip->source = NULL;
 	zip->info = NULL;
+	zip->zip64 = FALSE;
 	zip->vdir = NULL;
 	zip->stream = NULL;
 	zip->restlen = 0;
