@@ -547,6 +547,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 
 		for (i = 0 ; i < n ; i++) {
 			GValue *v;
+			const char *data0 = *data;
 			d (g_print ("\t[%d] ", i););
 			v = msole_prop_parse (section, type, data, data_end);
 			if (v) {
@@ -556,6 +557,8 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 				}
 				g_free (v);
 			}
+			if (*data == data0)
+				break;
 		}
 
 		res = g_new0 (GValue, 1);
@@ -1006,6 +1009,7 @@ msole_prop_read (GsfInput *in,
 		gsize gslen;
 		char *name;
 		guint8 const *start = data;
+		guint8 const *end = start + (size - 4);
 
 		g_return_val_if_fail (section->dict == NULL, FALSE);
 
@@ -1015,12 +1019,14 @@ msole_prop_read (GsfInput *in,
 
 		d ({ g_print ("Dictionary = \n"); gsf_mem_dump (data-4, size); });
 		n = type;
-		for (j = 0 ; j < n ; j++) {
+		for (j = 0; j < n; j++) {
+			g_return_val_if_fail (end - data >= 8, FALSE);
+
 			id = GSF_LE_GET_GUINT32 (data);
 			len = GSF_LE_GET_GUINT32 (data + 4);
 
 			g_return_val_if_fail (len < 0x10000, FALSE);
-			g_return_val_if_fail (len <= size - (data - start), FALSE);
+			g_return_val_if_fail (len <= end - data + 8, FALSE);
 
 			gslen = 0;
 			name = g_convert_with_iconv (data + 8,
@@ -1049,7 +1055,7 @@ msole_prop_read (GsfInput *in,
 
 		name = g_strdup (msole_prop_id_to_gsf (section, props[i].id, &linked));
 		d (g_print (" @ %x %x = ", (unsigned)props[i].offset, (unsigned)size););
-		val = msole_prop_parse (section, type, &data, data + size);
+		val = msole_prop_parse (section, type, &data, data + size - 4);
 
 		if (NULL != name && NULL != val) {
 			if (linked) {
